(** Tableaux blocks.
 * Blocks contain a set of schemata and a constraint.
 * They are the main pieces of data that rules work with.
 *
 * @see <http://membres-liglab.imag.fr/aravantinos/Publis/tableaux2009/tab09.pdf> A Schemata Calculus For Propositional Logic.
 * @author Vincent Aravantinos
 *)

(** Type of blocks. *)
type t

(** {6 Constructor } *)

val of_schs_and_cstr : cstr:Constraint.t -> schs:Schema.t list -> t

(** {6 Selectors } *)

val schs_of  : t -> Schema.t list
val cstr_of  : t -> Constraint.t
val is_empty : t -> bool

(** {6 Printing} *)

val to_string : t -> string
val print     : t -> unit

(** {6 Comparisons }*)

val equal : t -> t -> bool

(** {6 Others } *)

(** @return [true] if some schema in the block contains an iteration. *)
val contains_iterations : t -> bool

(** @return [true] if the block contains only literals. *)
val contains_only_literals : t -> bool

(** @return the maximum upper bound among all iterations occurring in the schemata of the block, if any, returns [None] if no iteration. *)
val max_bound : t -> Indexes.t option

(** @return the same block with duplicated schemas removed *)
val dedup : t -> t

(** @return the same block up to a shift (see looping.mli) but normalized for more efficient retrieval in the set of lemmas. *)
val normalize : t -> t

(** {6 Traversal of a block for rule application } *)

(** See regstab.ml for a better understanding *)
type branch_update =
  |Replace_with_schema                of Schema.t
  |Replace_with_schemas               of Schema.t * Schema.t
  |Replace_with_schema_add_constraint of Schema.t * Constraint.t
  |Add_constraint                     of Constraint.t

type branching =
  |Go_on
  |Close
  |Update of branch_update
  |Branch of branch_update * branch_update

(** Apply the given rules while checking if closure applied *)
val apply_rules_with_closure : t ->
    stat:((Stats.t -> unit) -> unit) ->
      rules:(sch:Schema.t -> cstr:Constraint.t -> branching) ->
        k:(t -> Satisfiability.t) -> 
          Satisfiability.t
(** Apply the given rules without checking if closure applied *)
val apply_rules : 
  t ->
    stat:((Stats.t -> unit) -> unit) ->
      rules:(sch:Schema.t -> cstr:Constraint.t -> branching) ->
        k:(t -> Satisfiability.t) -> 
          Satisfiability.t

(* {6 Purity }*)
include Purity.Purable with type purable = t
val purity : t -> Purity.t

(* {6 Looping}*)
module LoopingMap : Looping.LoopingMap with type loopable = t
