open Softcore open Block open Schema open Constraint open Satisfiability open Verbose open Stats

(*
 * The following global (imperative) variable contains a set of blocks that have already been seen, called ``lemmas''.
 * This set will be inspected for looping.
 *)
let lemmas = Lemmas.empty

(*
 * Propositional rules apply as long as possible.
 * As soon as a schema is encountered that can be applied a rule, it is decomposed according to this rule.
 * prop_rules is then applied recursively on the new nodes (the recursion is done through continuations).
 * Closure rule is also applied as soon as possible.
 *)
let rec prop_rules ~info bck = 
  Block.apply_rules_with_closure ~stat:info.stat bck ~k:(pure_rule ~info) ~rules:(fun ~sch ~cstr ->
    match sch with
    |(`Op(`And,sch1,sch2)   : Schema.t) -> info.stat incr.conj  ; Update(Replace_with_schemas(sch1,sch2))
    |(`Op(`Or,sch1,sch2)    : Schema.t) -> info.stat incr.disj  ; Branch(Replace_with_schema sch1, Replace_with_schema sch2)
    |(`Op(`Xor,sch1,sch2)   : Schema.t) -> info.stat incr.xor   ; Branch(Replace_with_schemas(neg sch1,sch2), Replace_with_schemas(sch1,neg sch2))
    |(`Op(`Equiv,sch1,sch2) : Schema.t) -> info.stat incr.equ   ; Branch(Replace_with_schemas(neg sch1,neg sch2), Replace_with_schemas(sch1,sch2))
    |(`Op(`Imply,sch1,sch2) : Schema.t) -> info.stat incr.imp   ; Branch(Replace_with_schema(neg sch1), Replace_with_schema(sch2))
    |(`Bot                  : Schema.t) -> Close (* Note: statistics for this are counted inside apply_rules_with_closure *)
    |_                                  -> Go_on)

and pure_rule ~info bck =
  let bck' = Block.dedup bck in
  match Block.purify bck' ~cstr:(Block.cstr_of bck') ~pur:(Block.purity bck') with
  |Modified.Unmodified,bck'   -> trivial_rule ~info bck'
  |Modified.Modified,purified -> trivial_rule ~info purified

and trivial_rule ~info bck =
  if Block.is_empty bck then ModelFound (cstr_of bck) else only_literals_rule ~info bck

(* If a block contains only literals and closure rule did not apply then it is satisfiable
 * (not proved anywhere but trivial)
 *)
and only_literals_rule ~info bck = if Block.contains_only_literals bck then ModelFound (cstr_of bck) else looping_rule ~info bck

(* The looping rule checks if the current node loops on a previously seen node (``lemma''), in which case we can conclude (NoModelFound) *)
and looping_rule ~info bck =
  if Block.contains_iterations bck
  then (if Lemmas.loops_among lemmas bck then (info.stat incr.loop; NoModelFound) else add_lemma ~info bck)
  else add_lemma ~info bck

and add_lemma ~info bck = Lemmas.add lemmas bck; info.lemma (Lemmas.Info_lemmas.add bck); it_rules ~info bck

(* Iteration rules apply in a way similar to propositional rules *)
and it_rules ~info bck =
  match Block.max_bound bck with
  |None     -> constraint_rule ~info bck
  |Some max -> Block.apply_rules bck ~stat:info.stat ~k:(constraint_rule ~info) ~rules:(fun ~sch ~cstr ->
      match sch with
      |(`It(`And,var,t1,t2,sch) : Schema.t) when Indexes.equal t2 max (* should always be the case except with relaxed conditions *) ->
          info.stat (incr.it_conj ^^ update_min_shift t2);
          Branch(Replace_with_schema_add_constraint(unfold_and_once ~var ~sch ~t1 ~t2, bounds_constraint t1 Le t2),
            Add_constraint(bounds_constraint t1 Gt t2))
      |(`It(`Or,var,t1,t2,sch) : Schema.t) when Indexes.equal t2 max  ->
          info.stat (incr.it_disj ^^ update_min_shift t2);
          Update(Replace_with_schema_add_constraint(unfold_or_once ~var ~sch ~t1 ~t2, bounds_constraint t1 Le t2))
      |(`Bot : Schema.t)                                              -> Close (* Note: statistics for this are counted inside apply_rules *)
      |_                                                              -> Go_on)

(* The constraint rule closes the tableau if the current constraint is unsatisfiable. *)
and constraint_rule ~info bck = if Constraint.is_false (Block.cstr_of bck) then (info.stat incr.closed; NoModelFound) else prop_rules ~info bck

and regstab ?(cstr=Constraint.top) ?(verbose=Verbose.all_false) sch =
  let stats = Stats.new_counter ~upper_bound:(Schema.max_bound sch) in
  let info_lemmas = Lemmas.Info_lemmas.new_ () in
  let verbose_closures ~default = Verbose.to_funs verbose ~default ~stats ~lemmas:info_lemmas in
  let result = prop_rules ~info:(verbose_closures ~default:()) (Block.of_schs_and_cstr ~schs:[sch] ~cstr) in
  set.lemmas stats (Lemmas.number lemmas);
  result, verbose_closures ~default:""

