open Softcore open Schema

let non_regular_string       = "Error: the input schema is not regular, "
let more_than_one_string     = "more than one parameter was detected, bad parenthesing?"
let diff_lower_bounds_string = "two iterations with different lower bounds were detected."
let diff_upper_bounds_string = "two iterations with different upper bounds were detected."
let inside_iteration_string  = "the parameter occurs inside an iteration."

(* The main function, called with the input channel and options *)
let main ch ~dimacs ~cnf ~param =
  if ch = stdin then print_endline "Taking input from stdin.";
  let lexbuf  = Lexing.from_channel ch in
  (try
    let (sch,_) = Parser.main Lexer.token lexbuf in
    match Schema.is_regular sch with
    |`Yes _ -> (
      let f = Schema.compute_schema sch param in
      if dimacs 
      then Formula.print_dimacs f
      else (if cnf
        then Formula.print_cnf f
        else Formula.print f
      )
    )
    |`No x -> print_endline (non_regular_string ^ "\n" ^ (match x with 
      |Different_lower_bounds -> diff_lower_bounds_string
      |Different_upper_bounds -> diff_upper_bounds_string
      |More_than_one_param    -> more_than_one_string
      |Param_inside_iteration -> inside_iteration_string))
  with Failure s -> print_endline s);
  print_newline ();
  flush stdout

(* Handling of command-line options, calls main *)
let _ =
  let param  = ref None
  and file   = ref None
  and dimacs = ref true
  and cnf    = ref true in
  let options = Arg.align [
    "-cnf"  , Arg.Set(cnf)      , " Display in conjunctive normal form (useful only when -H is set)";
    "-D"    , Arg.Set  (dimacs),  " Display in DIMACS format (default)";
    "-H"    , Arg.Clear(dimacs),  " Display in a human readable format instead of DIMACS";
    "-param", Arg.Int(fun x -> param := Some x), "Use [-param n] to indicate that the parameter shall be given the value [n]. Mandatory.";
  ]
  and usage = "sch2cnf n [file]  - computes the instance of a schema for value n of the parameter\n"
  and call_main s = 
    match !file with
    |None    -> file := Some(s)
    |Some(_) -> failwith "Too many arguments."
  in
  Arg.parse options call_main usage;
  match !param,!file with
    |None,_          -> failwith "The value of the parameter is missing (use [-param n])."
    |Some(p),None    -> main ~dimacs:(!dimacs) ~cnf:(!cnf) ~param:p stdin
    |Some(p),Some(f) -> read_wrap ~f:(main ~dimacs:(!dimacs) ~cnf:(!cnf) ~param:p) f
