(** Propositional schemata.
 *
 * Includes propositional formulae:
 * P_1, ~P_1, P_1 /\ P_2, P_1 \/ P_2, etc.
 * (propositional variables must be indexed i.e. P,Q,R are forbidden, P_1,Q_1,R_1 are not).
 *
 * And of course iterated schemata:
 * /\i=1..n P_i, \/i=1..n P_i.
 * @author Vincent Aravantinos
 *)

(** Type of schemata. Private. *)
type t = private
  [< `Lit of Literal.t                                               (* * a literal                                                         *)
  |`Op  of [ `And | `Or | `Xor | `Equiv | `Imply ] * t * t         (* * a formula whose head symbol is a binary operator                  *)
  |`It  of [ `And | `Or ] * Variable.t * Indexes.t * Indexes.t * t (* * a formula whose head symbol is the iteration of a binary operator *)
  |`Top                                                            (* * the true formula                                                  *)
  |`Bot ]                                                          (* * the false formula                                                 *)

(** {6 Constructors} *)

val bot     : t
val top     : t
val pos_lit : Proposition.t -> Indexes.t -> t
val neg_lit : Proposition.t -> Indexes.t -> t
val conj    : t -> t -> t
val disj    : t -> t -> t
val xor     : t -> t -> t
val equiv   : t -> t -> t
val imply   : t -> t -> t
val it_conj : Variable.t -> Indexes.t -> Indexes.t -> t -> t
val it_disj : Variable.t -> Indexes.t -> Indexes.t -> t -> t

(** {6 Print stuff} *)

val to_string      : t -> string
val list_to_string : t list -> string
val print          : t -> unit

(** {6 Comparison} *)
val equal         : t -> t -> bool
val total_compare : t -> t -> int

(** {6 Others} *)

(**
 * @return the maximal bound of iterations contained in the schema, returns [None] if no iteration.
 * @param acc the maximum will be also checked against acc.
 * *)
val max_bound           : ?acc:Indexes.t option -> t -> Indexes.t option

(**
 * @return [true] if the schema contains iterations, [false] else.
 *)
val contains_iterations : t -> bool

(**
 * @return the negation of the schema.
 *)
val neg                 : t -> t

(**
 * @param var variable of the returned iteration.
 * @param t1 lower bound of the returned iteration.
 * @param t2 upper bound of the returned iteration.
 * @return an iterated conjunction whose last rank is unfolded.
 *)
val unfold_and_once : var:Variable.t -> sch:t -> t1:Indexes.t -> t2:Indexes.t -> t

val unfold_or_once : var:Variable.t -> sch:t -> t1:Indexes.t -> t2:Indexes.t -> t
(** Similar with disjunction. *)


(** {6 Regularity of schemata} *)

exception Non_regular

(**
 * @return the only parameter of the schema if any.
 * @raise Non_regular if the schema has more than one parameter.
 *)
val get_param : t -> Variable.t option

(**
 * A type stating why a schema is not regular.
 *)
type non_regular_arg = Different_upper_bounds | Different_lower_bounds | More_than_one_param | Param_inside_iteration

(**
 * @return [true] if the schema is regular.
 *)
val is_regular : t -> [ `Yes of (Indexes.t * Indexes.t) option | `No of non_regular_arg ]

(**
 * @return computes the propositional formula obtained by substituting the only parameter with a given number.
 * @raise Non_regular if the schema has more than one parameter.
 *)
val compute_schema       : t -> int -> Formula.t

(**
 * @return the first expression (bound or index), if any, that involves the parameter of the schema.
 * This is used to normalize lemmata.
 * In this context, the input may only be an iteration or a literal so "the first expression" is actually the only expression.
 *)
val expr_involving_param : t -> Indexes.t option

(** {6 Purity } *)
include Purity.Purable with type purable = t

(** {6 Substitution } *)

(** Standard substitution of parameters *)
include Substitution.Substituable with type substituable = t and type term = Indexes.t and type variable = Variable.t

(** "Propositional variable Substitution", a module to allow subtitution of propositional variables by other propositional variables. *)
module PropSubs : Substitution.Substituable with type substituable = t and type term = Proposition.t and type variable = Proposition.t

(** {6 Looping } *)
module LoopingMap : Looping.LoopingMap with type loopable = t
